"use strict";

var _core = require("babel-runtime/core-js")["default"];

module.exports = fastpath;
var regExCache = {
    array: /(.*?)\[(.*?)\]/, //looks for [<contents>],
    cond: /([^<>=]+)([<>=][=]*)([^<>=]+)/, //looks for a>b , a<b, a=b, a>=b, a<=b patterns
    math: /([^\+-]+)([\+-])([^\+-]+)/, //looks for a+b, a-b
    exp: /(.*?)\((.*?)\)/ //look for (<contents>)
};

var prop = function (name) {
    return function (obj) {
        return obj[name];
    };
};
var genProp = function (fn) {
    return function (obj) {
        return obj[fn(obj)];
    };
};

function pick(arr) {

    if (!arr.length) {
        return function () {
            return false;
        };
    }
    var obj = {};
    arr.forEach(function (entry) {
        return obj[entry] = true;
    });
    return function (entry, idx) {
        return idx in obj;
    };
}

function filter(fn) {
    return function doFilter(arr) {
        var result = [];
        if (typeof arr !== "object") {
            return result;
        }

        var current = undefined;
        _core.Object.keys(arr).forEach(function (i) {
            return (current = fn(arr[i], i)) ? result[result.length] = arr[i] : undefined;
        });
        return result;
    };
}

function pickRange(start, end, step) {
    var fn;

    if (step) {
        fn = filter(function (entry, idx) {
            return (idx === start || (idx - start) % step === 0) && idx < end;
        });
    }
    return function (arr) {
        return fn && fn(arr) || (end !== undefined ? arr.slice(start, end) : arr.slice(start));
    };
}

function mapAndFilter(fn) {
    return function doMapFilter(arr) {
        var result = [];
        if (typeof arr !== "object") {
            return result;
        }

        var current = undefined;
        _core.Object.keys(arr).forEach(function (i) {
            return (current = fn(arr[i], i)) !== undefined ? result[result.length] = current : undefined;
        });
        return result;
    };
}

function spread(fn) {
    return function doSpread(arr) {
        var result = [];
        var current = undefined;
        arr.forEach(function (entry) {
            return result = (current = fn(entry)) ? result.concat(current) : result;
        });
        return result;
    };
}

function keys() {
    return function getValues(obj) {
        var result = [];

        if (typeof obj !== "object") {
            return result;
        }

        var keys = _core.Object.keys(obj);
        keys.forEach(function (item, i) {
            return result[i] = obj[keys[i]];
        });
        return result;
    };
}

function findKey(name) {
    return function find(obj) {
        var result = [];

        _core.Object.keys(obj).forEach(function (key) {
            var value = obj[key];
            if (name === key || name === "*") {
                result[result.length] = value;
            }

            if (typeof value === "object") {
                result = result.concat(find(value));
            }
        });
        return result;
    };
}

function processVal(val) {
    var isProp = val.indexOf && val.indexOf("@") !== -1;

    if (isProp) {
        val = val.replace("@.", "");
    } else {
        try {
            val = val.replace(/'/g, "");
        } catch (e) {}
    }
    return {
        isProp: isProp,
        val: val
    };
}

function conditional(lhs, op, rhs) {
    lhs = processVal(lhs);
    rhs = processVal(rhs);
    return function doCheck(entry) {
        var lvalue = lhs.isProp ? entry[lhs.val] : lhs.val,
            rvalue = rhs.isProp ? entry[rhs.val] : rhs.val;

        if (typeof lvalue === "object" || typeof rvalue === "object") {
            return false;
        }

        switch (op) {
            case ">":
                return lvalue > rvalue;

            case ">=":
                return lvalue >= rvalue;

            case "<":
                return lvalue < rvalue;

            case "<=":
                return lvalue <= rvalue;

            case "=":
                return lvalue == rvalue; // jshint ignore:line

            default:
                return false;
        }
    };
}

function math(lhs, op, rhs) {
    lhs = processVal(lhs);
    rhs = processVal(rhs);

    return function doMath(entry) {
        var lvalue = lhs.isProp ? entry[lhs.val] : lhs.val,
            rvalue = rhs.isProp ? entry[rhs.val] : rhs.val;

        if (typeof lvalue === "object" || typeof rvalue === "object") {
            return NaN;
        }
        switch (op) {
            case "+":
                return lvalue + rvalue;

            case "-":
                return lvalue - rvalue;

            case "*":
                return lvalue * rvalue;

            case "/":
                return lvalue / rvalue;

            default:
                return NaN;
        }
    };
}

function handleArrArg(loc, processors) {
    var exp = regExCache.cond.exec(loc);
    if (exp) {
        processors.push(spread(filter(conditional(exp[1], exp[2], exp[3]))));
    } else {
        exp = regExCache.math.exec(loc);
        if (exp) {
            processors.push(mapAndFilter(genProp(math(exp[1], exp[2], exp[3]))));
        } else {
            processors.push(spread(filter(prop(loc.replace("@.", "")))));
        }
    }
}

function handleArr(_ref, doSpread) {
    var locator = _ref.locator;
    var processors = _ref.processors;
    var idx = _ref.idx;

    var exp = regExCache.array.exec(locator.replace(/\s/g, ""));
    if (!exp) {
        throw new Error("Invalid JSONPath");
    }

    if (doSpread) {
        processors.push(spread(findKey(exp[1])));
    } else {
        //could be a pattern like $[<content>]
        if (exp[1] !== "$") {
            processors.push(mapAndFilter(prop(exp[1])));
        }
    }

    var tmp = undefined;
    if (locator.indexOf("@") !== -1) {
        tmp = regExCache.exp.exec(exp[2]);
        if (!tmp) {
            throw new Error("Invalid JSONPath");
        }
        handleArrArg(tmp[2].replace(/\s/g, ""), processors);
    } else {
        if (exp[2] === "*") {
            //[*]
            processors.push(spread(keys()));
        } else if ((idx = exp[2].indexOf(":")) !== -1) {
            tmp = exp[2].split(":");
            if (tmp.length < 3) {
                //n: or :n pattern
                processors.push(spread(idx === 0 ? pickRange(0, parseInt(tmp[1])) : pickRange(parseInt(tmp[0]))));
            } else {
                //start:end:step
                processors.push(spread(pickRange(parseInt(tmp[0]), parseInt(tmp[1]), parseInt(tmp[2]))));
            }
        } else {
            //[n], [1,2,...,n]
            processors.push(spread(filter(pick(exp[2].split(",")))));
        }
    }
}

function stitchLocator(locators, currIdx) {
    var loc = locators[currIdx];
    var idx = currIdx + 1;

    do {
        loc += "." + locators[idx];
    } while (locators[idx++].indexOf("]") === -1 && idx < locators.length);

    return {
        loc: loc,
        idx: idx
    };
}

function process(pattern) {

    var locators = pattern.replace("$.", "").split("."),
        processors = [],
        idx = 0,
        loc,
        idx1,
        idx2,
        tmp;

    while (idx < locators.length) {

        loc = locators[idx];
        if (loc === "*") {
            // .* operator
            processors.push(spread(keys()));
        } else if (loc.length === 0) {
            // .. operator
            tmp = locators[++idx];
            idx1 = tmp.indexOf("[");
            idx2 = tmp.indexOf("]");
            if (idx1 !== -1) {
                if (idx2 === -1) {
                    tmp = stitchLocator(locators, idx);
                    idx = tmp.idx;
                    loc = tmp.loc;
                }
                handleArr({ locator: tmp.loc || tmp, processors: processors, idx: idx }, true);
            } else {
                processors.push(spread(findKey(tmp)));
            }
        } else if (loc.indexOf("[") !== -1) {
            if (loc.indexOf("]") === -1) {
                //has  [<pattern>]
                tmp = stitchLocator(locators, idx);
                idx = tmp.idx;
                loc = tmp.loc;
            }
            handleArr({ locator: loc, idx: idx, processors: processors });
        } else {
            //simple .prop ref
            processors.push(mapAndFilter(prop(loc)));
        }
        idx++;
    }
    return processors;
}

function fastpath(pattern) {

    return {
        pattern: pattern,
        evaluate: function (obj) {
            return process(pattern).reduce(function (prev, current) {
                return current(prev);
            }, [obj]);
        }
    };
}